/* Emperor - an X windows strategy game for 1-6 players
   Copyright (C) 1998 Frank Rustemeyer.
   This file is part of the Emperor package.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston,
   MA 02111-1307, USA.

   If you have changes and you want them to be included in a future
   release of Emperor, please e-mail them to mambres@mud.tap.de.

   This file implements the X windows visualisation of the game.

   98/11/30, Frank Rustemeyer <mambres@mud.tap.de>: Released.
   99/01/29, Frank Rustemeyer <mambres@mud.tap.de>: Added roads.
   99/01/29, Frank Rustemeyer <mambres@mud.tap.de>: Added language support.
   99/02/07, Frank Rustemeyer <mambres@mud.tap.de>: Added "Forces" Window.
   99/02/09, Frank Rustemeyer <mambres@mud.tap.de>: Included cities in the Forces Window
   99/04/05, Frank Rustemeyer <mambres@mud.tap.de>: Bugfix: Handle save games without units
*/

#include <xview/xview.h>
//...more xview...h's are included in empx.h
#include <xview/cms.h>
#include <X11/Xlib.h>

#include "emp.h"
#include "empx.h"
#include "empgame.h"
#include "../xvcl/xvcl.h"
#include "colour.h"
#include "land.xpm"
#include "flags.xpm"
#include "units.xpm"

#include "flag1.xpm"
#include "flag2.xpm"
#include "flag3.xpm"
#include "flag4.xpm"
#include "flag5.xpm"
#include "flag6.xpm"

#define DEBUG(x) {printf("%s", x);fflush(stdout);}

#include <string.h>
  // strcpy  in mainwndw::settitle,
  // strcat  in mainwndw::settitle,
  // strlen  in mainwndw::paint.
//#include <stdio.h> included in empgame.h
  // sprintf in mainwndw::settitle,

extern distributor * d;  // object that distributes windows messages
game *g;
Xv_singlecolor cols[] = COLORARRAY;

// **************************************************** mainwndw ****

mainwndw::mainwndw(void)
{   setupwndw * sw;
    menu *filemenu, *zugmenu;
    static char *fileopts[] = 
                { T_SAVE, T_LOAD, T_MAP, T_STATS, T_FORCES, T_EXIT };
    static char *zugopts[] = 
                { T_DONE, T_MINES, T_ROAD, T_COMPMOVE, T_STAY };

    create_frame( WINDOWX, WINDOWY, "Emperor");
    set_min_size(100, 140);

    CmdPanel = new panel(this);
    CmdPanel->set_height(25);

    canv = new canvas(this, 0, 25, NCOLORS, cols);
    canv->set_fixed_size(SX*20, SY*20);
    canv->add_scrollbars();

    landxpm = new pixmap(canv->query_display(), land);
    flagsxpm = new pixmap(canv->query_display(), flags);
    unitsxpm = new pixmask(canv->query_display(), units);

    canv->mouse_events();
    canv->keyboard_events();

    waiting = 0;
    selected = NULL;
    selectedpos.set(-1,-1);

    g = new game();
    g->mp = new map();

    d->addwindow(this, canv->paint_window());
            // register window in distributor object
    sw = new setupwndw;
       sw->run();
    delete sw;

    filemenu = new menu(this, 6, fileopts );
    filebutton = new button(CmdPanel, T_GAME);
    filebutton->add_menu(filemenu);
   
    zugmenu = new menu(this, 5, zugopts );
    zugbutton = new button(CmdPanel, T_MOVE);
    zugbutton->add_menu(zugmenu);

    nextbutton = new button(CmdPanel, "->");
    stoppbutton = new button(CmdPanel, T_STOP);
    infobutton = new button(CmdPanel, T_UNIT);

    showmoves = new textview(CmdPanel, "");

    openchooser = new filechooser(this, FILE_OPEN, ".*.emp");
    savechooser = new filechooser(this, FILE_SAVE, ".*.emp");

    netzzug = new infobox(this, T_WAITING);

    g->initcities();

    bird = NULL;
    ctyw = NULL;
    score = NULL;

    nextplayer();
    settitle();
    }

void mainwndw::menu_selected(char *item)
{   if (waiting) return;
    if (!strcmp(item, T_EXIT))
    {   delete landxpm;
        delete flagsxpm;
        delete unitsxpm;
        if (bird) unbird();
        if (ctyw) uncity();
        if (score) unscore();
        if (forces) unforces();
        xv_destroy_safe(get_handle());
        return;
        }
    if (!strcmp(item, T_LOAD))
    {   openchooser->activate();
        return;
        }
    if (!strcmp(item, T_SAVE))
    {   savechooser->activate();
        return;
        }
    if (!strcmp(item, T_MAP))
    {   if (bird) unbird();
        else 
        {   bird = new birdwndw(this);
            bird->invalidate();
            }
        return;
        }
    if (!strcmp(item, T_STATS))
    {   if (score) unscore();
        else score = new scorewndw(this);
        }
    if (!strcmp(item, T_FORCES))
    {   if (forces) unforces();
        else forces = new armywndw(this);
        }
    if (!strcmp(item, T_STAY))
    {   if (!selected) return;
        selected->fortify();
        select(NULL);
        return;
        }
    if (!strcmp(item, T_MINES))
    {   if (!selected) return;
        if (selected->id() != MINENLEGER) return;
        if (g->stadt(selectedpos)) return;
        g->minelegen(selectedpos);
        invalidatesquare(selectedpos);
        return;
        }
    if (!strcmp(item, T_ROAD))
    {   if (!selected) return;
        if (selected->id() != INFANTRIE) return;
        if (!(selected->restmoves())) return;
        if (!g->freiland(selectedpos)) return;
        (*(g->mp))[selectedpos] = 6; /* make road */
        selected->countmove();
        select(NULL);
        return;
        }
    if (!strcmp(item, T_DONE)) 
    {   nextplayer();
        return;
        }
    if (!strcmp(item, T_COMPMOVE))
    {   busy(1);
        g->compmove();
        select(NULL);
        findnextunit();
        busy(0);
        invalidate(FULL);
        if (forces) forces->update();
        return;
        }
    }       

void mainwndw::button_pressed(Xv_object item)
{   if (waiting) return;
    
    if (item == nextbutton -> get_handle())
    {   if (!nextunit) return;  // in case no unit available at all
        select(nextunit);
        setwindow(nextunit->getpos());
        findnextunit();
        return;
        }

    if (item == stoppbutton -> get_handle())
    {   if (!selected) return;
        if (!selected->restmoves()) return;
        selected->nomoves();
        if (g->stadt(selectedpos))
        {    messagebox *mb;
             char msg[120];

             mb = new messagebox(this, selected->repair(msg));
             delete mb;
             }
        //proceed: switch to next unit
        findnextunit();
        if (!nextunit)
        {   stoppbutton->enable(0);
            invalidatemoves();
            }
        button_pressed(nextbutton->get_handle());
        return;
        }

     if (item == infobutton -> get_handle())
     {  if (!selected) return;
        if ((selected->id() == TRANSPORTER) && ((transporter *) selected)->getnextunload())
        {   unloadwndw *dialog;
            dialog = new unloadwndw(this, (transporter *) selected);
            dialog->run();
            delete dialog;
            }
         else 
         {   char s[40];
             messagebox *mb;
             mb = new messagebox(this, selected->hitpointstring(s));
             delete mb;
             }
         return;
         }
    }

void mainwndw::file_open(char *file)
{   messagebox *mb;
    netwndw* nw;
    char text[60];
    int i;
 
    if (ctyw) uncity();

//    nw = new netwndw();
//      nw->run();
//    delete nw;  
   
    select(NULL);

    g->activeplayer = 0;
    invalidate(FULL);

    if (!g->load(file)) 
    {   mb = new messagebox(this, T_ERRORLOADING);
        delete mb;
        }
    else
    {   unit *u;

        sprintf(text, T_PLAYERSTURN, g->sp[g->activeplayer].name);
        mb = new messagebox(this, text);
        delete mb;
 
        u = (unit *) FIRSTMEMBER(g->units);
        if ((u) && (u != (unit *) (u->last()->first())))
        {   mb = new messagebox(this, T_DATAERROR);
            delete mb;
            }

        if (g->sp[g->activeplayer].ext)
        {   waiting = 1;
            i = g->activeplayer;
            do
               if (++i > 6) i = 1;
            while (!g->sp[i].used || g->sp[i].ext);
            g->activeplayer = i;
            netzzug->show(1);
            busy(1);
//            SetTimer(hW, TIMER, 5000, NULL);
            nextbutton->enable(0);
            stoppbutton->enable(0);
            }
         else
         {   findnextunit();
             button_pressed(nextbutton->get_handle());
             }
         settitle();
         g->lookout(g->activeplayer);
         invalidate(FULL);
         if (forces) forces->update();
         }
    }

void mainwndw::file_save(char *file)
{   if (!g->save(file))
    {   messagebox *mb;
        mb = new messagebox(this, T_ERRORSAVING);
        delete mb;
        }
    }

mainwndw::~mainwndw(void)
{      
//    For safety reasons, those are better done before exiting
//    the main loop (in menu_selected)
//    delete landxpm;
//    delete flagsxpm;
//    delete unitsxpm;
    }

void mainwndw::invalidate(int mode)
{   canv->invalidate();
    if (bird) bird->invalidate();
    } 

void mainwndw::settitle(void)
{   char t[100];
    if (g->activeplayer)
    {   strcpy(t, "Emperor: ");
	strcat(t, g->sp[g->activeplayer].name);
	}
    else strcpy(t, "Emperor");
    xv_set((Frame) get_handle(), FRAME_LABEL, t, END);
    }

void mainwndw::invalidatesquare(coordinate c)
{   static Xv_xrectlist r;
    Xv_Window w;
    w = canv->paint_window();

    if (!g->activeplayer) return;
    if (!ONMAP(c)) return;
    r.count = 1;
    r.rect_array[0].x = c.x * 20 -2;
    r.rect_array[0].y = c.y * 20 -2;
    r.rect_array[0].height = r.rect_array[0].width = 22;

    paint(canv->get_handle(), w, canv->query_display(),
          xv_get(w, XV_XID), &r);
    }

void mainwndw::invalidatearea(coordinate c1, coordinate c2)
{   static Xv_xrectlist r;
    Xv_Window w;
    int x, y;
    w = canv->paint_window();

    if (!g->activeplayer) return;
    r.count = 1;
    x = ((c1.x > c2.x) ? c2.x : c1.x) * 20;
    if (x >= 20) x -= 20;
    r.rect_array[0].x = x;
    y = ((c1.y > c2.y) ? c2.y : c1.y) * 20;
    if (y >= 20) y -= 20;
    r.rect_array[0].y = y;
    r.rect_array[0].width = 
	((c1.x > c2.x) ? (c1.x - c2.x) : (c2.x - c1.x))*20 + 40;
    r.rect_array[0].height = 
	((c1.y > c2.y) ? (c1.y - c2.y) : (c2.y - c1.y))*20 + 40;

    paint(canv->get_handle(), w, canv->query_display(),
	    xv_get(w, XV_XID), &r);
    }

void mainwndw::setwindow(coordinate c)
{   if (canv->visible(c.x * 20, c.y * 20)) return;
    canv->centre(c.x * 20 + 10, c.y * 20 + 10);
    }

void mainwndw::invalidatemoves(void)
{   char str[50], rstr[8], hstr[30];

    if (selected) 
    {    sprintf(str, T_UNITMOVES, 
             selected->hitpointstring(hstr), selected->restmovestring(rstr));
         if (selected->id() == TRANSPORTER) 
            if (((transporter *) selected)->getnextunload())
                 strcat(str, T_FULL); 
            else strcat(str, T_EMPTY);
         }
    else strcpy(str, "");

    showmoves->set_value(str);
    }

void mainwndw::findnextunit(void)
{   class unit *current;
    int start = 0;

    current = (unit *) FIRSTMEMBER(g->units);
    if (!current) nextunit = NULL;
    else
    {   if (selected)
	{   start = selected->getlfdnr();
	    current = (unit *) selected->next();
	    if (!current) current = (unit *) selected->first();
            }
	nextunit = current -> findnext(g->activeplayer, start);
        }
    nextbutton->enable(nextunit != NULL);
    }

void mainwndw::select(unit *u)
{   coordinate c;
    c.set(selectedpos.x, selectedpos.y);
    selectedpos.set(-1,-1);
    selected = u;
    invalidatesquare(c);
    stoppbutton->enable(selected && selected->restmoves());
    if (!selected) 
    {   selectedpos.set(0,0);
	invalidatemoves();
	infobutton->enable(0);
	//        festbutton->enable(0);
	return;
        }
    //    festbutton->enable(1);
    infobutton->enable(1);
    selected->unfortify();
    selectedpos = selected->getpos();

    g->unitmap[selectedpos] = selected->id();
    //    minebutton->enable((selected->id() == MINENLEGER)
    //        && !g->stadt(selectedpos) && !g->minebekannt(selectedpos));
    invalidatesquare(selectedpos);
    invalidatemoves();
    }

void mainwndw::selectcity(city *ct)
{   select(NULL);
    selectedpos = ct->getpos();
    setwindow(selectedpos);
    }

void mainwndw::nextplayer(void)
{   char msg[100];
    char i;
    char newgame;
    int inplay;

    uncity();
//    unscore();
    select(NULL);

    i = g->activeplayer;
    newgame = !i;
    do
	if (++i > 6) i = 1;
    while (!g->sp[i].used);

    g->activeplayer = 0;
    settitle();

    invalidate(FULL);

    if (g->sp[i].ext)
    {   char flagfile[20];
	FILE *f;
	waiting = 1;
	if (!newgame)
	{   messagebox *mb;
	    g->activeplayer = i-1;
	    while (!g->save("network.emp"))
	    {   mb = new messagebox(this, T_ERRORNETSAVE);
		delete mb;
                }
	    sprintf(flagfile, "turn%d.flg", i);
	    while (!(f = fopen(flagfile, "w")))
	    {   mb = new messagebox(this, T_ERRORFLAGSAVE);
		delete mb;
                }
	    fprintf(f, "Ready to go !\n");
	    fclose(f);
	    }
	do
	    if (++i > 6) i = 1;
	while (!g->sp[i].used || g->sp[i].ext);
	g->activeplayer = i;
	netzzug->show(1);
	busy(1);
	//        SetTimer(hW, TIMER, 5000, NULL);
	nextbutton->enable(0);
	//       minebutton->enable(0);
	stoppbutton->enable(0);
	return;
        }

    inplay =  ((city *) FIRSTMEMBER(g->cities)) -> produce(i); // Production
    if (FIRSTMEMBER (g->units))
        inplay += ((unit *) FIRSTMEMBER(g->units)) -> clearmoves(i); // set move counter to 0

    g->lookout(i);
    if (forces) forces->update();

    if (g->sp[i].computer)
    {   busy(1);
	g->activeplayer = i;
	g->compmove();
	busy(0);
	nextplayer();
        }
    else
    {   messagebox *mb;
	strcpy(msg, g->sp[i].name);
        if (!inplay) strcat(msg, T_ELIMINATED);
        else strcat(msg, T_YOURTURN);
	mb = new messagebox(this, msg);
	delete mb;
	g->activeplayer = i;
	settitle();
	findnextunit();
	invalidate(MAP);
	button_pressed(nextbutton->get_handle());
        }
   }

void mainwndw::paint(Canvas cv, Xv_Window x, Display *dsp, 
	Window wnd, Xv_xrectlist *xr)
{   GC gc;
    int c;

    coordinate i, min, max, n;
    char p;
    Frame hW;

    hW = (Frame) get_handle();

    if (!xr)  
    {   min.set(0,0);
        max.set(20*SX, 20*SY);
        }
    else 
    {   min.set(20*SX, 20*SY);
        max.set(0,0);
        for (c = 0; c < xr->count; c++)
        {   if (xr->rect_array[c].x < min.x) min.x = xr->rect_array[c].x;
            if (xr->rect_array[c].y < min.y) min.y = xr->rect_array[c].y;
            if (xr->rect_array[c].x + xr->rect_array[c].width > max.x)
                    max.x = xr->rect_array[c].x + xr->rect_array[c].width;
            if (xr->rect_array[c].y + xr->rect_array[c].height > max.y)
                    max.y = xr->rect_array[c].y + xr->rect_array[c].height;
            }
        }

    min.set(min.x - (min.x % 20), min.y - (min.y % 20));

    p = g->activeplayer;

    gc = DefaultGC(dsp, DefaultScreen(dsp));

    XSetLineAttributes(dsp, gc, 2, LineSolid, CapNotLast, JoinMiter);

    if (p && !waiting)
    {   for(i.x = min.x / 20; 20*i.x <= max.x ; i.x++)
            for(i.y = min.y / 20; 20*i.y <= max.y; i.y++)
            {   if (ONMAP(i))
		    if (g->bekannt(i))
                    {   if (g->strasse(i)) 
                        {   /* display land square as background */
                            landxpm->display(dsp, wnd, 0, 60, 20, 20,
                                             20*i.x, 20*i.y);
                            XSetForeground(dsp, gc, canv->colour(GREY));
                            c = 1;
                            LOOP(n, -1, -1, 1, 1, 1) 
                               if (n.x || n.y)
                               {   if ((g->roadmap)[i] & c)
                                       XDrawLine(dsp, wnd, gc, 
                                                20*i.x+10, 20*i.y+10,
                                                20*i.x+10 + n.x*10,
                                                20*i.y+10 + n.y*10);
                                   c *= 2;
                                   }
                            }
                        else
                        {   landxpm->display(dsp, wnd, 0, 20*((*(g->mp))[i]), 
                                             20, 20, 20*i.x, 20*i.y);
                            }
                        if(g->minebekannt(i))
                        {    XSetBackground(dsp, gc, canv->colour(WHITE));
                             XSetForeground(dsp, gc, canv->colour(WHITE));
                             XFillRectangle(dsp, wnd, gc, 20*i.x+7, 20*i.y+7,
                                            6, 6);
                             }
                        if(g->beobachtet(i) && g->ownership[i])
                             if (g->meins(i) || (g->unitmap[i] != UBOOT) || g->stadt(i))
                                  flagsxpm->display(dsp, wnd, 0, 6*(g->ownership[i]-1),
                                                    10, 6, 20*i.x+10, 20*i.y);
			if(g->beobachtet(i) && g->unitmap[i] 
                           && (g->meins(i) || (g->unitmap[i] != UBOOT)))
                             unitsxpm->display(dsp, wnd, 0, 20*(g->unitmap[i]-1),
                                               20, 20, 20*i.x, 20*i.y);
			 }
		    else
 		    {	 XSetForeground(dsp, gc, canv->colour(BLACK));
                         XSetBackground(dsp, gc, canv->colour(BLACK));
                         XFillRectangle(dsp, wnd, gc, 20*i.x, 20*i.y,
                                        20, 20);
			 }
          	}
		       
        XSetBackground(dsp, gc, canv->colour(GREY));

        if (selected)
        {   char str[8];
            XSetForeground(dsp, gc, canv->colour(WHITE));
            XDrawRectangle(dsp, wnd, gc, (selectedpos.x)*20,
                                         (selectedpos.y)*20, 19, 19);
            }
        }
    else   /* Busy... clear window */
    {   XSetForeground(dsp, gc, canv->colour(GREY));
        XSetBackground(dsp, gc, canv->colour(GREY));
        XFillRectangle(dsp, wnd, gc, 0, 0, 20 * SX, 20 * SY);
        }
    }

void mainwndw::unbird(void)
{   if (!bird) return;
    delete bird;
    bird = NULL;
    }
  
void mainwndw::unscore(void)
{   if (!score) return;
    delete score;
    score = NULL;
    }

void mainwndw::uncity(void)
{   if (!ctyw) return;
    delete ctyw;
    ctyw = NULL;
    }

void mainwndw::unforces(void)
{   if (!forces) return;
    delete forces;
    forces = NULL;
    }

int mainwndw::make_selection(coordinate c)
{   unit *u;

    if (!ONMAP(c)) return 0;
    if (!(g->bekannt(c))) return 0;
    u = (unit *) g->units.find(c);
    if (!u) return 0;
    if (u->getowner() != g->activeplayer) return 0;
    select(u);
    findnextunit();
    return 1;
    }

void mainwndw::make_move(coordinate moveto)
{   coordinate oldpos;
 
    if (!selected) return;
    if (!ONMAP(moveto)) return;
    if (selectedpos == moveto) return;

    oldpos = selectedpos;
    moveto = selected->movetowards(moveto);
    selected = selected->alive();

    if (selected)
    {   selectedpos = selected->getpos();
        if (!ONMAP(selectedpos)) selected = NULL; 
        }
    select(selected);

    invalidatearea(oldpos, moveto);
    invalidatemoves();
    findnextunit();
    if (forces) forces->update();
    }

void mainwndw::event(Xv_Window wnd, Event *e,
                     Notify_arg arg, Notify_event_type t)
{   coordinate click;
    city *c;

    click.set(event_x(e)/20, event_y(e)/20);

    if (event_is_ascii(e) && event_is_up(e)) // ignore key downs
       return;

    switch(event_action(e))
    {    case ACTION_SELECT:

              if (event_is_down(e))
              {   make_selection(click);
                  break;
                  }
              
              if (!selected) break;
              make_move(click);
              break;

         case ACTION_MENU:

              if (!event_is_up(e)) break;

              if (!ONMAP(click)) break;
	      if (!(g->bekannt(click))) break;

	      c = (city *) g->cities.find(click);
              if (c)
              {   if (!ctyw)
                  {    ctyw = new citywndw(this, c);
                       }
                  else
                       ctyw->setcity(c); // pass c to city
                  break;
                  }
              if (!make_selection(click)) break;
              button_pressed(infobutton->get_handle());
              break;

// [Freshman] hier steht noch alter Windows-Code. Der soll
//            eigentlich bewirken, dass man mit den Pfeiltasten
//            den Bildschirmausschnitt bewegen kann (wie mit 
//            den Scrollbars. Da ich hier die Scrollbars nicht
//            von Hand behandeln muss, gibt es die Funktion
//            reactscroll() nicht mehr; aber mit
//            canv->moveto(coordinate x) kann man den Ausschnitt
//            bewegen.

//                case VK_LEFT:
//                      reactscroll(SB_LINEUP, 0, SB_HORZ, &(n.x));
//                      sbh->update(n.x);
//                      return 0;
//                 case VK_RIGHT:
//                      reactscroll(SB_LINEDOWN, 0, SB_HORZ, &(n.x));
//                      sbh->update(n.x);
//                      return 0;
//                 case VK_UP:
//                      reactscroll(SB_LINEUP, 0, SB_VERT, &(n.y));
//                      sbv->update(n.y);
//                      return 0;
//                 case VK_DOWN:
//                      reactscroll(SB_LINEDOWN, 0, SB_VERT, &(n.y));
//                      sbv->update(n.y);
//                      return 0;

// [Freshman] Das hier funktioniert prima mit den "echten Zahlen"
//            Es ist aber eigentlich fuer die NumPad-Tasten gedacht
//            (zum Bewegen der Einheiten)

//         case '7':
         case ACTION_GO_DOCUMENT_START:
              click.set(-1, -1); 
              make_move(selectedpos + click);
              break;
//         case '8':
         case 'k':
         case ACTION_GO_COLUMN_BACKWARD:
              click.set(0, -1);
              make_move(selectedpos + click);
              break;
//         case '9':
         case ACTION_GO_PAGE_BACKWARD:
              click.set(1, -1);
              make_move(selectedpos + click);
              break;
//         case '4':
         case 'h':
         case ACTION_GO_CHAR_BACKWARD:
              click.set(-1, 0);
              make_move(selectedpos + click);
              break;
//         case '6':
         case 'l':
         case ACTION_GO_CHAR_FORWARD:
              click.set(1, 0);
              make_move(selectedpos + click);
              break;
//         case '1':
         case ACTION_GO_DOCUMENT_END:
              click.set(-1, 1);
              make_move(selectedpos + click);
              break;
//         case '2':
         case 'j':
         case ACTION_GO_COLUMN_FORWARD:
              click.set(0, 1);
              make_move(selectedpos + click);
              break;
//         case '3':
         case ACTION_GO_PAGE_FORWARD:
              click.set(1, 1);
              make_move(selectedpos + click);
              break;

         case 'i':
              button_pressed(infobutton->get_handle());
              break;
         case 'f':
              menu_selected(T_STAY);
              break;
         case 'm':
              menu_selected(T_MINES);
              break;
         case 'r':
              menu_selected(T_ROAD);
              break;

         case '+':
         case 9: // TAB
//                 case VK_NUMPAD5:
              button_pressed(nextbutton->get_handle());
              break;

//         case END: 
         case 13: // RETURN
              menu_selected(T_DONE);
              break;

         case '-':
         case 32: // space

              button_pressed(stoppbutton->get_handle());
              break;

         default:
              if (event_is_iso(e)) 
                  fprintf(stderr, "Key %c (%d)\n", event_action(e),
                                    event_action(e));
              return;
         }
     }

/*
long mainwndw::wndproc(WNDPROCTAIL)
{ 
    switch (msg)
    { 
        case WM_TIMER:

             {   FILE *f;
                 char flagfile[20];
                 sprintf(flagfile, "turn%d.flg", g->activeplayer);
                 f = fopen(flagfile, "r");
                 if (!f) break;
                 fclose(f);
                 KillTimer(hW, TIMER);
                 waiting = 0;
                 DestroyWindow(info);
                 while (!g->load("network.emp"))
                     MB(hW, "Cannot load network file... Try again.", "Error", MBERROR);
                 while(remove("network.emp"))
                     MB(hW, "Cannot clear network file... Try again.", "Error", MBERROR);
                 while(remove(flagfile))
                     MB(hW, "Cannot remove network file... Try again.", "Error", MBERROR);
                 nextplayer();
                 }
             break;
    }
*/
// **************************************************** citywndw ****

citywndw::citywndw(window *o, city *c) 
{   int i;
 
    create_frame(160, 200, "Stadt");
    set_min_size(160, 200);
    set_max_size(160, 200);
    owner = o; 
    pan = new panel(this);

    okbutton = new button(pan, "OK");
    setbutton = new button(pan, T_CHANGE);
    besitzer = new textview(pan, "");
    besitzer->set_pos(10, 40);
    progress = new gauge(pan);
    progress->set_pos(10, 70);
    progress->set_width(120);

    unitbox = new listbox(pan, NULL, 4);
    unitbox->set_pos(10,100);
    unitbox->add_string(T_NOPRODUCTION);
    unitbox->add_string(T_ARMY);
    unitbox->add_string(T_TANK);
    unitbox->add_string(T_AIRCRAFT);
    unitbox->add_string(T_SUBMARINE);
    unitbox->add_string(T_BOAT);
    unitbox->add_string(T_FRIGATE);
    unitbox->add_string(T_BATTLESHIP);
    unitbox->add_string(T_FERRY);
    unitbox->add_string(T_MINEBOAT);
           
    setcity(c);
    show(1);
    }

citywndw::~citywndw(void)
{   xv_destroy_safe(get_handle());
    delete okbutton;
    delete setbutton;
    delete besitzer;
    delete unitbox;
    delete progress;
    delete pan;
    }

void citywndw::setcity(city *c)
{   char bes[100];

    ct = c;

    strcpy(bes, T_OWNER);
    if (!g->beobachtet(ct->getpos()))
       strcat(bes, T_UNKNOWN);
    else
       strcat(bes, g->sp[ct->getowner()].name);
    besitzer->set_value(bes);

    if (ct->getowner() != g->activeplayer)
    {   setbutton->enable(0);
        progress->show(0);
        return;
        }
     
    setbutton->enable(1);

    if (ct->produced)
    {    progress->set_range(0, ct->produced->prodtime());
         progress->set_value(ct->prodtime());
         progress->show(1);
         unitbox->set_value(ct->produced->id());
         }
    else 
    {    progress->show(0);
         unitbox->set_value(0);
         }
    }

void citywndw::button_pressed(Xv_object item)
{   int v;
    if (item == okbutton->get_handle())
    {   ((mainwndw *) owner) -> uncity();
        return;
        }
    ct->setprod(unitbox->get_value());
    setcity(ct);
    }

// *************************************************** scorewndw ****

scorewndw::scorewndw(window *o)
{   int i;
    
    create_frame(285, 220, T_STATISTICS);
    set_min_size(285, 220);
    set_max_size(285, 220);
    owner = o;
   
    pan = new panel(this);

    citybn = new button(pan, T_CITIES);
    areabn = new button(pan, T_EXPLORED);
    armybn = new button(pan, T_ARMIES);

    for (i = 0; i < 6; i++)
    {   score[i] = new gauge(pan);
        score[i] -> set_pos(55, 40+30*i);
        score[i] -> set_width(200);
        } 

    flagge[0] = new imageview(pan, flag1);
    flagge[0] -> set_pos(20, 40);
    flagge[1] = new imageview(pan, flag2);
    flagge[1] -> set_pos(20, 70);
    flagge[2] = new imageview(pan, flag3);
    flagge[2] -> set_pos(20, 100);
    flagge[3] = new imageview(pan, flag4);
    flagge[3] -> set_pos(20, 130);
    flagge[4] = new imageview(pan, flag5);
    flagge[4] -> set_pos(20, 160);
    flagge[5] = new imageview(pan, flag6);
    flagge[5] -> set_pos(20, 190);

    button_pressed(citybn->get_handle());
    show(1);
    }

scorewndw::~scorewndw(void)
{   int i;
    show(0);
    xv_destroy_safe(get_handle());
    delete citybn;
    delete areabn;
    delete armybn;
    delete pan;
    for (i = 0; i < 6; i++)  
    {   delete score[i];
        delete flagge[i];
        }
    }

void scorewndw::countcities(void)
{   city * run;
    int i, max;

    citybn->enable(FALSE);
    areabn->enable(TRUE);
    armybn->enable(TRUE);

    for(i=1; i<7; i++) sc[i] = 0;
    max = 0;
    
    run = (city *) FIRSTMEMBER(g->cities);

    while(run)
    {   max++;
        sc[run->getowner()]++;
        run = (city *) run->next();
        }
    for(i=0; i<6; i++)
        score[i] -> set_range(0, max);
    }

void scorewndw::countarea(void)
{   coordinate c;
    int i;

    citybn->enable(TRUE);
    armybn->enable(TRUE);
    areabn->enable(FALSE);
    for(i=1; i<7; i++) sc[i] = 0;
    
    LOOP(c, 0, 0, SX-1, SY-1, 1)
        for(i=1; i<7; i++)
            if (g->bekannt(c, i)) sc[i]++;

    for(i=1; i<7; i++)
    {   score[i-1] -> set_range(0, 100);
        sc[i] = ((long int) sc[i] * 100) / (SX * SY);
        }
    }


void scorewndw::countarmies(void)
{   unit *u;
    int i, max;

    citybn->enable(TRUE);
    areabn->enable(TRUE);
    armybn->enable(FALSE);

    for(i=1; i<7; i++) sc[i] = 0;
    max = 0;
    
    u = (unit *) FIRSTMEMBER(g->units);

    while(u)
    {   max += u->gethp();
        sc[u->getowner()] += u->gethp();
        u = (unit *) u->next();
        }
    for(i=1; i<7; i++)
    {   sc[i] = ((long int) sc[i] * 100) / max;
        score[i-1] -> set_range(0, 100);
        }
    }

void scorewndw::button_pressed(Xv_object item)
{   int i;

    if (item == citybn->get_handle())
        countcities();
    else if (item == areabn->get_handle())
        countarea();
    else countarmies();

    for(i=1; i<7; i++)
        if (sc[i])
        {    score[i-1] -> set_value(sc[i]);
             score[i-1] -> show(1);
             }
        else score[i-1] -> show(0);
    }

// *************************************************** armywndw ****

armywndw::armywndw(window *o)
{   int i;
    char **unit_data;
    
    create_frame(455, 220, T_FORCES);
    set_min_size(455, 220);
    set_max_size(455, 220);
    owner = o;
   
    pan = new panel(this);

    okbutton = new button(pan, "OK");
    okbutton -> set_pos(5, 10);

    flagge[0] = new imageview(pan, flag1);
    flagge[0] -> set_pos(20, 50);
    flagge[1] = new imageview(pan, flag2);
    flagge[1] -> set_pos(20, 80);
    flagge[2] = new imageview(pan, flag3);
    flagge[2] -> set_pos(20, 110);
    flagge[3] = new imageview(pan, flag4);
    flagge[3] -> set_pos(20, 140);
    flagge[4] = new imageview(pan, flag5);
    flagge[4] -> set_pos(20, 170);
    flagge[5] = new imageview(pan, flag6);
    flagge[5] -> set_pos(20, 200);

    for (i = 1; i <= UNIT_TYPES; i++)
    {   unit_data = extract_unit(units, i);
        einheit[i-1] = new imageview(pan, unit_data);
        free_array(unit_data);
        einheit[i-1] -> set_pos(60 + 40 * (i-1), 10);
        }

    for(i=0; i<6*UNIT_TYPES; i++)
    {   zahl[i] = new textview(pan, "-");
        zahl[i] -> set_pos(60 + 40 * (i % UNIT_TYPES), 50 + 30 * (i / UNIT_TYPES));
        }

    unit_data = extract_land(land, 6);
    stadtpic = new imageview(pan, unit_data);
    free_array(unit_data);
    stadtpic -> set_pos(420, 10);

    for(i = 0; i < 6; i++)
    {   stadt[i] = new textview(pan, "-");
        stadt[i] -> set_pos(420, 50 + 30 * i);
        }

    update();
    show(1);
    }

armywndw::~armywndw(void)
{   int i;
    show(0);
    xv_destroy_safe(get_handle());
    delete okbutton;
    delete pan;
    for (i = 0; i < 6; i++)  
       delete flagge[i];
    for (i = 0; i < UNIT_TYPES; i++)
       delete einheit[i];
    for (i = 0; i < 6 * UNIT_TYPES; i++)
       delete zahl[i]; 
    delete stadtpic;
    for (i = 0; i < 6; i++)
       delete stadt[i];
    }

void armywndw::update(void)
{   int i;
    unit *u;
    city *run;
    char c[10];
    int sc[6 * UNIT_TYPES];

    for(i=0; i < 6*UNIT_TYPES; i++) sc[i] = 0;

    u = (unit *) FIRSTMEMBER(g->units);

    while(u)
    {   sc[(u->getowner()-1)* UNIT_TYPES + (u->id() -1)] ++;
        u = (unit *) u->next();
        }

    for(i = 0; i < 6*UNIT_TYPES; i++)
        if (!sc[i])
           zahl[i] -> set_value("-");
        else
        {  sprintf(c, "%d", sc[i]);
           zahl[i] -> set_value(c);
           }

    for(i=1; i<7; i++) sc[i] = 0;
    
    run = (city *) FIRSTMEMBER(g->cities);

    while(run)
    {   sc[run->getowner()]++;
        run = (city *) run->next();
        }
    for(i=0; i<6; i++)
        if (!sc[i+1])
            stadt[i] -> set_value("-");
        else
        {   sprintf(c, "%d", sc[i+1]);
            stadt[i] -> set_value(c);
            }
    }
           
void armywndw::button_pressed(Xv_object item)
{   int i;

    if (item == okbutton->get_handle())
       ((mainwndw *) owner) -> unforces();
    }

char ** armywndw::extract_unit(char ** pixdata, int unittype)
{   char **result;
    int i;

    result = (char **) malloc(24 * sizeof(char *));
    for(i = 0; i < 4; i++)
    {   result[i] = (char *) malloc(30 * sizeof(char));
        strcpy(result[i], pixdata[i]);
        }
    strcpy(result[0], "20 20 3 1"); 
    strcpy(result[1], "       c white");
    for(i = 0; i < 20; i++)
    {   result[4+i] = (char *) malloc(25 * sizeof(char));
        strcpy(result[4+i], pixdata[4 + 20 * (unittype -1) + i]);
        }
    return result;
    }

char ** armywndw::extract_land(char ** pixdata, int landtype)
{   char **result;
    int i;

    result = (char **) malloc(31 * sizeof(char *));
    for(i = 0; i < 11; i++)
    {   result[i] = (char *) malloc(30 * sizeof(char));
        strcpy(result[i], pixdata[i]);
        }
    strcpy(result[0], "20 20 10 1"); 
    for(i = 0; i < 20; i++)
    {   result[11+i] = (char *) malloc(25 * sizeof(char));
        strcpy(result[11+i], pixdata[11 + 20 * (landtype -1) + i]);
        }
    return result;
    }

void armywndw::free_array(char ** array)
{   int i;
    for (i = 0; i < 24; i++) free (array[i]);
    free (array);
    }    

// **************************************************** birdwndw ****

birdwndw::birdwndw(window *o) 
{   create_frame(3*SX, 3*SY, T_MAP);
    set_min_size(3*SX, 3*SY);
    set_max_size(3*SX, 3*SY);
    owner = o; 

    canv = new canvas(this, 0, 0, NCOLORS, cols);
    canv->mouse_events();

//  [Freshman] In order to register additional events for that
//             window, use this:
//
//  xv_set(canv->paint_window(),  
//         WIN_EVENT_PROC, centralevent,
//         WIN_CONSUME_EVENTS, [register event here], END,
//         END);

    d->addwindow(this, canv->paint_window());
            // register window in distributor object
    notify_destroy();
    show(1);
    }

birdwndw::~birdwndw(void)
{   show(0);
    xv_destroy_safe(get_handle());
    delete canv;
    }

void birdwndw::destroy(void)
{   delete this;
    }

void birdwndw::invalidate(void)
{   canv->invalidate();
    }

void birdwndw::paint(Canvas cv, Xv_Window x, Display *dsp, 
	Window wnd, Xv_xrectlist *xr)
{   GC gc;
    Frame hW;
    coordinate i;

    hW = (Frame) get_handle();

    gc = DefaultGC(dsp, DefaultScreen(dsp));

    if (g->activeplayer)
    {    LOOP(i, 0,0, SX-1, SY-1, 1)
             if (g->bekannt(i))
             {   switch((*(g->mp))[i])
                 {   case 0:
                     case 1:
                          XSetBackground(dsp, gc, canv->colour(BLUE));
                          XSetForeground(dsp, gc, canv->colour(BLUE));
                          break;
                     case 2:
                          XSetBackground(dsp, gc, canv->colour(LIGHTBLUE));
                          XSetForeground(dsp, gc, canv->colour(LIGHTBLUE));
                          break;
                     case 3:
                     case 4:
                          XSetBackground(dsp, gc, canv->colour(GREEN));
                          XSetForeground(dsp, gc, canv->colour(GREEN));
                          break;
                     case 5:
                          XSetBackground(dsp, gc, canv->colour(RED));
                          XSetForeground(dsp, gc, canv->colour(RED));
                          break;
                     case 6:
                          XSetBackground(dsp, gc, canv->colour(BLACK));
                          XSetForeground(dsp, gc, canv->colour(BLACK));
                          break;
                     }
                 XFillRectangle(dsp, wnd, gc, 3*i.x, 3*i.y, 3, 3);
                 }
         }
    else 
    {   XSetBackground(dsp, gc, canv->colour(GREY));
        XSetForeground(dsp, gc, canv->colour(GREY));
        XFillRectangle(dsp, wnd, gc, 0, 0, 3*SX, 3*SY);
        }
    }

// [Freshman] This is where the events go
//            Currently, only mouseclicks are used.
void birdwndw::event(Xv_Window wnd, Event *e,
                     Notify_arg arg, Notify_event_type t)
{   coordinate click;

    click.set(event_x(e)/3, event_y(e)/3);

    if((event_action(e) == ACTION_SELECT) && event_is_up(e))
        ((mainwndw *) owner)->setwindow(click);
    }
